@@ -36,7 +36,9 @@ __all__ = ['GcmMode']
 
 from binascii import unhexlify
 
-from Crypto.Util.py3compat import byte_string, bord, _copy_bytes
+from Crypto.Util.py3compat import bord, _copy_bytes
+
+from Crypto.Util._raw_api import is_buffer
 
 from Crypto.Util.number import long_to_bytes, bytes_to_long
 from Crypto.Hash import BLAKE2s
@@ -70,7 +72,7 @@ def _build_impl(lib, postfix):
         imp_funcs = [ getattr(lib, x + "_" + postfix) for x in funcs ]
     except AttributeError:      # Make sphinx stop complaining with its mocklib
         imp_funcs = [ None ] * 3
-    params = dict(list(zip(funcs, imp_funcs)))
+    params = dict(zip(funcs, imp_funcs))
     return GHASH_Imp(**params)
 
 
@@ -184,8 +186,9 @@ class GcmMode(object):
 
         if len(nonce) == 0:
             raise ValueError("Nonce cannot be empty")
-        if isinstance(nonce, str):
-            raise TypeError("Nonce must be a byte string")
+
+        if not is_buffer(nonce):
+            raise TypeError("Nonce must be bytes, bytearray or memoryview")
 
         # See NIST SP 800 38D, 5.2.1.1
         if len(nonce) > 2**64 - 1:
@@ -204,8 +207,8 @@ class GcmMode(object):
             raise ValueError("Parameter 'mac_len' must be in the range 4..16")
 
         # Allowed transitions after initialization
-        self._next = [self.update, self.encrypt, self.decrypt,
-                      self.digest, self.verify]
+        self._next = ["update", "encrypt", "decrypt",
+                      "digest", "verify"]
 
         self._no_more_assoc_data = False
 
@@ -222,21 +225,19 @@ class GcmMode(object):
                                   **cipher_params
                                   ).encrypt(b'\x00' * 16)
 
-        # Step 2 - Compute J0 (integer, not byte string!)
+        # Step 2 - Compute J0
         if len(self.nonce) == 12:
-            self._j0 = bytes_to_long(self.nonce + b"\x00\x00\x00\x01")
+            j0 = self.nonce + b"\x00\x00\x00\x01"
         else:
-            fill = (16 - (len(nonce) % 16)) % 16 + 8
+            fill = (16 - (len(self.nonce) % 16)) % 16 + 8
             ghash_in = (self.nonce +
                         b'\x00' * fill +
-                        long_to_bytes(8 * len(nonce), 8))
-            self._j0 = bytes_to_long(_GHASH(hash_subkey, ghash_c)
-                                     .update(ghash_in)
-                                     .digest())
+                        long_to_bytes(8 * len(self.nonce), 8))
+            j0 = _GHASH(hash_subkey, ghash_c).update(ghash_in).digest()
 
         # Step 3 - Prepare GCTR cipher for encryption/decryption
-        nonce_ctr = long_to_bytes(self._j0 >> 32, 12)
-        iv_ctr = (self._j0 + 1) & 0xFFFFFFFF
+        nonce_ctr = j0[:12]
+        iv_ctr = (bytes_to_long(j0) + 1) & 0xFFFFFFFF
         self._cipher = factory.new(key,
                                    self._factory.MODE_CTR,
                                    initial_value=iv_ctr,
@@ -249,7 +250,7 @@ class GcmMode(object):
         # Step 6 - Prepare GCTR cipher for GMAC
         self._tag_cipher = factory.new(key,
                                        self._factory.MODE_CTR,
-                                       initial_value=self._j0,
+                                       initial_value=j0,
                                        nonce=b"",
                                        **cipher_params)
 
@@ -281,12 +282,12 @@ class GcmMode(object):
             A piece of associated data. There are no restrictions on its size.
         """
 
-        if self.update not in self._next:
+        if "update" not in self._next:
             raise TypeError("update() can only be called"
                             " immediately after initialization")
 
-        self._next = [self.update, self.encrypt, self.decrypt,
-                      self.digest, self.verify]
+        self._next = ["update", "encrypt", "decrypt",
+                      "digest", "verify"]
 
         self._update(assoc_data)
         self._auth_len += len(assoc_data)
@@ -363,10 +364,10 @@ class GcmMode(object):
           Otherwise, ``None``.
         """
 
-        if self.encrypt not in self._next:
+        if "encrypt" not in self._next:
             raise TypeError("encrypt() can only be called after"
                             " initialization or an update()")
-        self._next = [self.encrypt, self.digest]
+        self._next = ["encrypt", "digest"]
 
         ciphertext = self._cipher.encrypt(plaintext, output=output)
 
@@ -416,10 +417,10 @@ class GcmMode(object):
           Otherwise, ``None``.
         """
 
-        if self.decrypt not in self._next:
+        if "decrypt" not in self._next:
             raise TypeError("decrypt() can only be called"
                             " after initialization or an update()")
-        self._next = [self.decrypt, self.verify]
+        self._next = ["decrypt", "verify"]
 
         if self._status == MacStatus.PROCESSING_AUTH_DATA:
             self._pad_cache_and_update()
@@ -441,10 +442,10 @@ class GcmMode(object):
         :Return: the MAC, as a byte string.
         """
 
-        if self.digest not in self._next:
+        if "digest" not in self._next:
             raise TypeError("digest() cannot be called when decrypting"
                             " or validating a message")
-        self._next = [self.digest]
+        self._next = ["digest"]
 
         return self._compute_mac()
 
@@ -491,10 +492,10 @@ class GcmMode(object):
             or the key is incorrect.
         """
 
-        if self.verify not in self._next:
+        if "verify" not in self._next:
             raise TypeError("verify() cannot be called"
                             " when encrypting a message")
-        self._next = [self.verify]
+        self._next = ["verify"]
 
         secret = get_random_bytes(16)
 
